# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

import logging
import struct
import socket
import errno
import datetime
import time
import binascii
import threading
import collections

import iu_config as config
import iu_db as db



def GetObjectAttribute(obj, name):
  try:
    return getattr(obj, name)
  except AttributeError:
    return None



class ServiceId:
  FILE_REPUTATION   = 1
  HOST_REPUTATION   = 4
  HASH_INFO         = 11
  HOST_INFO         = 12
  URL_INFO          = 13


class RequestFlags:
  Json = 1


class ReplyStatus:
  Success   = 0
  TempError = 1
  PermError = 2
  NoData    = 3


class Request(object):
  def __init__(self, service_type, key):
    self.service_type = service_type
    self.serviceId = service_type.serviceId
    self.key = key
    self.replyQueue = None

    self.requestId = None
    self.replyFlags = None
    self.replyStatus = None
    self.result = None

  def __str__(self):
    return '%s(%s)' % (self.service_type.name, self.key)

  def GetServiceId(self):
    return self.serviceId

  def Success(self):
    return self.replyStatus == ReplyStatus.Success or self.replyStatus == ReplyStatus.NoData

  def NoData(self):
    return self.replyStatus == ReplyStatus.NoData

  def GetCommRequestFlags(self):
    return self.service_type.GetCommRequestFlags(self)

  def GetCommRequestData(self):
    return self.service_type.GetCommRequestData(self)

  def SetCommReplyData(self,reply):
    return self.service_type.SetCommReplyData(self,reply)


def Hex2Bin(s):
  pfx = s[0:2]
  if pfx=='0x' or pfx=='0X':
    s = s[2:]
  return binascii.a2b_hex(s)

def Bin2Hex(b):
  return binascii.b2a_hex(b).lower()

def IpIs6(ip):
  if str(ip).find(':') == -1:
    return False
  else:
    return True

if hasattr(socket,'inet_pton') and hasattr(socket,'inet_ntop'):

  def Ip2Bin(ip):
    ip = str(ip)
    if IpIs6(ip):
      return socket.inet_pton(socket.AF_INET6,ip)
    else:
      return socket.inet_pton(socket.AF_INET,ip)

  def Bin2Ip(ipb):
    if len(ipb) != 4:
      return socket.inet_ntop(socket.AF_INET6,ipb)
    else:
      return socket.inet_ntop(socket.AF_INET,ipb)

else:

  def Ip2Bin(ip):
    ip = str(ip)
    if IpIs6(ip):
      raise Exception('No IP6 address support for this platform')
    else:
      return socket.inet_aton(ip)

  def Bin2Ip(ipb):
    if len(ipb) != 4:
      raise Exception('No IP6 address support for this platform')
    else:
      return socket.inet_ntoa(ipb)

def IsIpAddress(addr):
  try:
    Ip2Bin(addr)
  except Exception:
    return False
  return True


def TimeFromKsn(s):
  try:
    if s == None or len(s) == 0:
      return None
    s = s.replace('T',' ')
    s = s.replace('Z',' ')
    try:
      tm = datetime.datetime.strptime(s[0:19],'%Y-%m-%d %H:%M:%S')
    except Exception:
      tm = datetime.datetime.strptime(s[0:10],'%Y-%m-%d')
  except Exception as e:
    logging.warning('Time (%s) conversion error: %s' % (s, e))
    return None

  if tm.year <= 1970:
    return None
  return tm

def TimeToSql(tm):
  if tm == None:
    return None
  return tm.strftime('%Y-%m-%d %H:%M:%S')

def TimeFromKsnToSql(s):
  return TimeToSql(TimeFromKsn(s))

def CurrentTimeUtc():
  return datetime.datetime.utcnow()

def CurrentTimeLocal():
  return datetime.datetime.now()

def TimeDeltaToSeconds(timedelta):
  try:
    return 24*60*60*timedelta.days + timedelta.seconds + 1e-6*timedelta.microseconds
  except AttributeError:
    return timedelta


class Queue(object):
  def __init__(self, maxSize=0):
    self._maxSize = maxSize or 0
    self._size = 0
    self._lock = threading.RLock()
    self._rcond = threading.Condition(self._lock)
    self._wcond = threading.Condition(self._lock)
    self._queue = collections.deque()
    self._closed = False

  def Put(self, obj):
    with self._lock:
      while 1:
        if self._closed:
          raise EOFError()
        if self._maxSize != 0 and self._size == self._maxSize:
          self._wcond.wait()
        else:
          self._queue.appendleft(obj)
          self._size += 1
          self._rcond.notify()
          return

  def Get(self):
    with self._lock:
      while 1:
        try:
          obj = self._queue.pop()
          self._size -= 1
          if self._maxSize != 0:
            self._wcond.notify()
          return obj
        except IndexError:
          if self._closed:
            raise EOFError()
        self._rcond.wait()

  def Close(self):
    with self._lock:
      self._closed = True
      self._rcond.notify_all()
      self._wcond.notify_all()

  def IsEmpty(self):
    with self._lock:
      if self._queue:
        return False
    return True



lastActiveTime = time.time()

def UpdateActivity():
  global lastActiveTime
  lastActiveTime = time.time()

def WaitForIdle(delayToWait, watchdogProc = None):
  global lastActiveTime
  while 1:
    currentTime = time.time()
    currentDelay = currentTime - lastActiveTime
    if currentDelay > delayToWait:
      break
    logging.debug('Processing idle for %s (max delay: %s secs)' % (currentDelay, delayToWait))
    time.sleep(1)




class RequestMap:

  def __init__(self):
    self.lock = threading.Lock()
    self.requestMap = {}
    self.nextRequestId = 1

  def Put(self,req):
    with self.lock:
      requestId = self.nextRequestId
      self.nextRequestId += 1
      if (self.nextRequestId == 0):
        raise Exception('Request ID overflow')
      self.requestMap[requestId] = req
      return requestId

  def Get(self,requestId):
    with self.lock:
      req = None
      if requestId in self.requestMap:
        req = self.requestMap[requestId]
        del self.requestMap[requestId]
      return req

  def LogMissingReplies(self):
    with self.lock:
      for requestId in self.requestMap:
        req = self.requestMap[requestId]
        logging.warning('Missing reply for request %s: %s' % (requestId, req))

  def IsEmpty(self):
    with self.lock:
      for requestId in self.requestMap:
        return False
    return True


def recvall(sk,size):
  buf = ''
  while size > 0:
    try:
      portion = sk.recv(size)
      buf += portion
      size -= len(portion)
    except socket.error as e:
      if e.args[0] == errno.EAGAIN:
        continue
      raise
    if len(portion) == 0:
      raise EOFError('Server has closed the socket')
  return buf


_ReqHdrPack = struct.Struct('<3sBIII')
_RspHdrPack = struct.Struct('<3sBIII')

_REQ_SIG = 'AR1'
_RSP_SIG = 'AP1'

_REQ_HDR_SIZE = _ReqHdrPack.size
_RSP_HDR_SIZE = _RspHdrPack.size


class CommChannel(object):
  def __init__(self):
    self._requestMap = RequestMap()
    self._socket = None
    self._opened = False
    self._closeLock = threading.Lock()

  def SendRequest(self, req):
    try:
      flags = req.GetCommRequestFlags()
      data  = req.GetCommRequestData()
    except Exception as e:
      logging.error('CommRequest: invalid request data: %s, error: %s' % (req,e))
      req.service_type.OnRecvErr()
      return
    req.requestId = self._requestMap.Put(req)
    logging.debug('CommRequest: id:%s service:%s flags:%s size:%s data:%s' % (req.requestId, req.serviceId, flags, len(data), req))
    req.service_type.OnSend()
    pkt = _ReqHdrPack.pack(_REQ_SIG,flags,req.requestId,req.serviceId,len(data)) + data
    self._socket.sendall(pkt)

  def ReceiveReply(self):
    while True:
      hdr = recvall(self._socket,_RSP_HDR_SIZE)
      hf = _RspHdrPack.unpack_from(hdr, 0)
      if hf[0] != _RSP_SIG:
        raise IOError('Protocol error: invalid reply header')

      logging.debug('CommReply: id:%s status:%s flags:%s size:%s' % (hf[2],hf[3],hf[1],hf[4]))

      replySize = hf[4]
      replyData = recvall(self._socket,replySize)

      requestId = hf[2]
      reply = self._requestMap.Get(requestId)
      if reply == None:
        logging.warning('CommReply: request ID not matched: %s' % (requestId,))
        continue

      reply.replyFlags = hf[1]
      if ((reply.replyFlags & RequestFlags.Json) != 0):
        reply.replyJson = replyData

      reply.replyStatus = hf[3]

      if reply.Success():
        try:
          if len(replyData) == 0:
            reply.replyStatus = ReplyStatus.NoData
          if reply.replyStatus == ReplyStatus.Success:
            reply.SetCommReplyData(replyData)
          reply.service_type.OnRecv()
        except Exception as e:
          logging.error('CommReply: parse error: %s svc:%s id:%s' % (e,reply.serviceId, requestId))
          logging.error('CommReply: failed reply data:%s' % (replyData,))
          reply.replyStatus = ReplyStatus.PermError
          reply.service_type.OnRecvErr()
      else:
        reply.service_type.OnRecvErr()

      logging.debug('CommReply received: %s status:%s' % (reply,reply.replyStatus))
      return reply

  def Connect(self, host=config.APT_HOST, port=config.APT_PORT):
    port = int(port)
    logging.info('Connecting to %s:%s...' % (host,port))
    try:
      self._socket = socket.socket()
      self._socket.settimeout(TimeDeltaToSeconds(config.APT_KSN_TIMEOUT))
      self._socket.connect((host, port))
      logging.info('Connected to %s:%s' % (host,port))
      self._opened = True
    except Exception as e:
      logging.info('Cannot connect to %s:%s: %s' % (host,port,e))
      raise

  def Disconnect(self):
    with self._closeLock:
      if self._opened:
        logging.info('Disconnecting socket')
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()
        self._opened = False

  def HasUnprocessedRequests(self):
    if not self._requestMap.IsEmpty():
      return True
    return False

  def LogMissingReplies(self):
    self._requestMap.LogMissingReplies()



class TestCommServer(object):
  def __init__(self):
    self._listenSocket = None
    self._workSocket = None
    self._processor = None
    self._worker = None

  def Start(self, portNumber, processor):
    logging.info('Starting test server on port %s' % (portNumber,))

    try:
      self._processor = processor
      self._worker = threading.Thread(name='TestSrv', target=self._RunServer)

      self._listenSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      self._listenSocket.bind(('127.0.0.1', portNumber))
      self._listenSocket.listen(1)

      self._worker.start()
    except Exception as e:
      logging.error('Error starting test server: %s' % (e,))

  def Stop(self):
    self.CloseSockets()
    if self._worker != None:
      try:
        self._worker.join()
      except Exception:
        pass
      self._worker = None

  def CloseSockets(self):
    if self._listenSocket != None:
      try:
        self._listenSocket.close()
      except Exception:
        pass
      self._listenSocket = None
    if self._workSocket != None:
      try:
        self._workSocket.close()
      except Exception:
        pass
      self._workSocket = None

  def _RunServer(self):
    logging.info('Running test server')
    try:
      while True:
        logging.info('Test server: waiting for connections')
        self._workSocket, addr = self._listenSocket.accept()
        logging.info('Test server: accepted connection from %s' % (addr,))

        self.ProcessRequests()

        logging.info('Test server: disconnected')

    except Exception as e:
      logging.info('Stopping test server: %s' % (e,))
      self.CloseSockets()

  def ProcessRequests(self):
    while True:
      hdr = recvall(self._workSocket,_REQ_HDR_SIZE)
      hf = _ReqHdrPack.unpack_from(hdr, 0)
      if hf[0] != _REQ_SIG:
        raise IOError('Protocol error: invalid request header')

      reqSize = hf[4]
      reqData = recvall(self._workSocket,reqSize)

      requestId = hf[2]
      serviceId = hf[3]

      replyStatus, replyData = self._processor(serviceId, reqData)
      flags = RequestFlags.Json

      pkt = _RspHdrPack.pack(_RSP_SIG,flags,requestId,replyStatus,len(replyData)) + replyData
      self._workSocket.sendall(pkt)


class CommInfoSource(object):
  def __init__(self, name, parms):
    self._name = name
    self._commChannel = CommChannel()
    self._sendThread = None
    self._recvThread = None
    self._host = config.APT_HOST
    self._port = config.APT_PORT
    self._requestQueue = Queue(parms['queue_size'])
    self._started = False
    self._last_send_time = 0

  def Start(self):
    self._commChannel.Connect(self._host, self._port)
    self._recvThread = threading.Thread(name=('%s_recv' % (self._name,)), target=self._DoReceiveReplies)
    self._recvThread.daemon = True
    self._sendThread = threading.Thread(name=('%s_send' % (self._name,)), target=self._DoSendRequests)
    self._sendThread.daemon = True
    self._recvThread.start()
    self._sendThread.start()
    self._started = True

  def Stop(self):
    self._requestQueue.Close()
    if (self._sendThread != None):
      self._sendThread.join()
      self._sendThread = None
    try:
      self._commChannel.Disconnect()
    except Exception:
      pass
    if (self._recvThread != None):
      self._recvThread.join()
      self._recvThread = None
    self._started = False

  def SendRequest(self, req):
    if not self._started:
      raise Exception('CommInfoSource stopped')
    self._requestQueue.Put(req)

  def _DoSendRequests(self):
    try:
      logging.info('Sender thread started')
      while 1:
        try:
          req = self._requestQueue.Get()
        except EOFError:
          logging.info('Send queue closed')
          break

        sleep_time = self._last_send_time + config.APT_KSN_REQUEST_INTERVAL - time.time()
        if sleep_time > 0:
          time.sleep(sleep_time)
        self._last_send_time = time.time()

        self._commChannel.SendRequest(req)

    except socket.timeout:
      logging.error('Error sending request: socket timeout')
      self._started = False
    except Exception as e:
      logging.error('Error sending request: %s' % (e,))
    finally:
      try:
        self._commChannel.Disconnect()
      except Exception as e:
        logging.info('Disconnect (send): %s' % (e,))
      finally:
        self._requestQueue.Close()
      logging.info('Sender thread stopped')

  def _DoReceiveReplies(self):
    try:
      logging.info('Receiver thread started')
      while 1:
        try:
          reply = self._commChannel.ReceiveReply()
        except (EOFError, socket.timeout):
          logging.info('Server connection closed')
          break
        except Exception as e:
          logging.error('Error receiving reply: %s' % (e,))
          break
        try:
          reply.replyTime = CurrentTimeUtc()
          if (reply.replyQueue == None):
            logging.warn('Reply has no target queue')
          else:
            reply.replyQueue.Put(reply)
        except Exception as e:
          logging.info('Error processing reply (%s): %s' % (reply,e))
    except Exception as e:
      logging.error('Exception in receiver thread: %s' % (e,))
    finally:
      try:
        self._commChannel.Disconnect()
      except Exception as e:
        logging.info('Disconnect (recv): %s' % (e,))
      finally:
        self._requestQueue.Close()
        self._started = False
      try:
        self._commChannel.LogMissingReplies()
      except Exception:
        pass
      logging.info('Receiver thread stopped')

  def HasUnprocessedRequests(self):
    if self._commChannel.HasUnprocessedRequests():
      return True
    if not self._requestQueue.IsEmpty():
      return True
    return False


def DbConnection():
  return db.AptDbConnection()

def DbName(schemaName,objectName):
  if db.UseSchemas and schemaName!=None and schemaName!='':
    return '%s.%s' % (schemaName,objectName)
  else:
    return objectName

def DbBinary(s):
  return db.Binary(s)


def DbBoolean(s):
  return db.Boolean(s)

DbOpt = db.SqlOpt

def HashToDb(hashStr):
  if db.UseBinaryHash:
    return DbBinary(Hex2Bin(hashStr))
  else:
    return str(hashStr)

def HashFromDb(hashVal):
  if db.UseBinaryHash:
    return Bin2Hex(hashVal)
  else:
    return str(hashVal)

UseInetAddr = db.UseInetAddr



class InfoProvider(object):
  def __init__(self):
    self._started = False

    logging.info('Creating InfoProvider')
    self.commSource = CommInfoSource('Comm', config.APT_COMM_PARMS)

    self._infoSources = []
    self._infoSources.append(self.commSource)

    self._serviceMap = {}
    self._serviceMap[ServiceId.FILE_REPUTATION] = self.commSource
    self._serviceMap[ServiceId.HOST_REPUTATION] = self.commSource
    self._serviceMap[ServiceId.HASH_INFO] = self.commSource
    self._serviceMap[ServiceId.HOST_INFO] = self.commSource
    self._serviceMap[ServiceId.URL_INFO] = self.commSource

  def Start(self):
    try:
      logging.info('Starting InfoProvider...')
      for source in self._infoSources:
        source.Start()
      self._started = True
      logging.info('InfoProvider started')
    except Exception as e:
      logging.error('Error starting InfoProvider: %s' % (e,))
      raise

  def Stop(self):
    logging.info('Stopping InfoProvider...')
    for source in self._infoSources:
      try:
        source.Stop()
      except Exception:
        pass
    self._started = False
    logging.info('InfoProvider stopped')

  def SendRequest(self, request, replyQueue):
    logging.info('Processing request: %s' % (request,))
    if not self._started:
      raise Exception('Error processing request: InfoProvider not started')

    serviceId = GetObjectAttribute(request, 'serviceId')
    if serviceId not in self._serviceMap:
      raise Exception('Invalid serviceId: %s' % (serviceId,))

    source = self._serviceMap[request.serviceId]
    if not source._started:
      raise socket.error('CommInfoSource stopped')

    try:
      request.replyQueue = replyQueue
      source.SendRequest(request)
    except Exception as e:
      logging.error('Error processing request: %s' % (e,))
      raise

  def HasUnprocessedRequests(self):
    return any(source.HasUnprocessedRequests() for source in self._infoSources)
